@@ -50,7 +50,7 @@ from typing import (
     Tuple, Type)
 
 from . import _streamer as _sync_streamer
-from ._streamer import is_exception, _default_peek_func, MAX_THREADS
+from ._streamer import is_exception, _default_peek_func, EventUpstreamer, MAX_THREADS
 
 
 logger = logging.getLogger(__name__)
@@ -70,10 +70,10 @@ class Stream(collections.abc.AsyncIterator, _sync_streamer.StreamMixin):
 
     def __init__(self, instream: Union[Stream, Iterator, Iterable]):
         if isinstance(instream, Stream):
-            self._to_shutdown = instream._to_shutdown
+            self._crashed = instream._crashed
             self._instream = instream
         else:
-            self._to_shutdown = threading.Event()
+            self._crashed = EventUpstreamer()
             if hasattr(instream, '__anext__'):
                 self._instream = instream
             else:
@@ -115,15 +115,9 @@ class Stream(collections.abc.AsyncIterator, _sync_streamer.StreamMixin):
         return await self._instream.__anext__()
 
     async def __anext__(self):
-        try:
-            z = await self._get_next()
-            self.index += 1
-            return z
-        except StopAsyncIteration:
-            raise
-        except:
-            self._to_shutdown.set()
-            raise
+        z = await self._get_next()
+        self.index += 1
+        return z
 
     async def collect(self) -> list:
         return [v async for v in self]
@@ -148,8 +142,8 @@ class Stream(collections.abc.AsyncIterator, _sync_streamer.StreamMixin):
     def drop_if(self, func: Callable[[int, T], bool]) -> Dropper:
         return Dropper(self, func)
 
-    def head(self, n: int) -> Head:
-        return Head(self, n)
+    def head(self, n: int) -> Header:
+        return Header(self, n)
 
     def peek(self, func: Callable[[int, T], None] = None) -> Peeker:
         if func is None:
@@ -168,6 +162,7 @@ class Stream(collections.abc.AsyncIterator, _sync_streamer.StreamMixin):
                   *,
                   workers: Optional[Union[int, str]] = None,
                   return_exceptions: bool = False,
+                  keep_order: bool = None,
                   **func_args) -> Union[Transformer, ConcurrentTransformer]:
         '''
         When `workers = N > 0`, the worker threads (if `func` is sync)
@@ -186,7 +181,9 @@ class Stream(collections.abc.AsyncIterator, _sync_streamer.StreamMixin):
             1 <= workers <= 100
         return ConcurrentTransformer(
             self, func, workers=workers,
-            return_exceptions=return_exceptions)
+            return_exceptions=return_exceptions,
+            keep_order=keep_order,
+        )
 
 
 class Batcher(Stream):
@@ -237,7 +234,7 @@ class Dropper(Stream):
             return z
 
 
-class Head(Stream):
+class Header(Stream):
     def __init__(self, instream: Stream, n: int):
         super().__init__(instream)
         assert n >= 0
@@ -265,9 +262,9 @@ class IterQueue(asyncio.Queue, collections.abc.AsyncIterator):
     PUT_SLEEP = 0.00045
     NO_MORE_DATA = object()
 
-    def __init__(self, maxsize: int, to_shutdown: threading.Event):
+    def __init__(self, maxsize: int, downstream_crashed: EventUpstreamer):
         super().__init__(maxsize)
-        self._to_shutdown = to_shutdown
+        self._downstream_crashed = downstream_crashed
 
     async def put_end(self):
         await self.put(self.NO_MORE_DATA)
@@ -278,7 +275,7 @@ class IterQueue(asyncio.Queue, collections.abc.AsyncIterator):
                 super().put_nowait(x)
                 break
             except asyncio.QueueFull:
-                if self._to_shutdown.is_set():
+                if self._downstream_crashed.is_set():
                     return
                 await asyncio.sleep(self.PUT_SLEEP)
 
@@ -290,6 +287,8 @@ class IterQueue(asyncio.Queue, collections.abc.AsyncIterator):
                     raise StopAsyncIteration
                 return z
             except asyncio.QueueEmpty:
+                if self._downstream_crashed.is_set():
+                    raise StopAsyncIteration
                 await asyncio.sleep(self.GET_SLEEP)
 
 
@@ -298,41 +297,35 @@ class Buffer(Stream):
         super().__init__(instream)
         assert 1 <= maxsize <= 10_000
         self.maxsize = maxsize
-        self._q = IterQueue(maxsize, self._to_shutdown)
-        self._err = None
+        self._q = IterQueue(maxsize, self._crashed)
+        self._upstream_err = None
         self._task = None
         self._start()
 
     def _start(self):
-        async def foo(instream, q):
+        async def foo(instream, q, crashed):
             try:
                 async for v in instream:
-                    await q.put(v)
-                    if self._to_shutdown.is_set():
+                    if crashed.is_set():
                         break
+                    await q.put(v)
                 await q.put_end()
             except Exception as e:
                 # This should be exception while
                 # getting data from `instream`,
                 # not exception in the current object.
-                self._err = e
-                self._to_shutdown.set()
+                self._upstream_err = e
+                await q.put_end()
 
-        self._task = asyncio.create_task(foo(self._instream, self._q))
-
-    def _stop(self):
-        if self._task is not None:
-            _ = self._task.result()
-            self._task = None
-
-    def __del__(self):
-        self._stop()
+        self._task = asyncio.create_task(foo(self._instream, self._q, self._crashed))
 
     async def _get_next(self):
-        if self._err is not None:
-            self._stop()
-            raise self._err
-        return await self._q.__anext__()
+        try:
+            return await self._q.__anext__()
+        except StopAsyncIteration:
+            if self._upstream_err is not None:
+                raise self._upstream_err
+            raise
 
 
 def is_async(func):
@@ -367,6 +360,7 @@ class Transformer(Stream):
         except Exception as e:
             if self.return_exceptions:
                 return e
+            self._crashed.set()
             raise
 
 
@@ -377,65 +371,89 @@ class ConcurrentTransformer(Stream):
                  *,
                  workers: int,
                  return_exceptions: bool = False,
+                 keep_order: bool = None,
                  ):
         assert workers >= 1
         super().__init__(instream)
         self.func = func
         self.workers = workers
         self.return_exceptions = return_exceptions
-        self._err = []
+        self.keep_order = workers > 1 and (keep_order is None or keep_order)
+        self._upstream_err = []
         self._tasks = []
 
         if is_async(func):
             self._async = True
-            self._outstream = IterQueue(workers * 8, self._to_shutdown)
+            self._outstream = IterQueue(workers * 1024, self._crashed)
             self._start_async()
         else:
             self._async = False
-            self._outstream = _sync_streamer.IterQueue(
-                workers * 8, self._to_shutdown)
+            self._outstream = _sync_streamer.IterQueue(1024, self._crashed)
             self._start_sync()
 
     def _start_async(self):
-        async def _process(in_stream, out_stream, func,
-                           lock, finished, return_exceptions):
-            while not finished.is_set():
+        async def _process(in_stream, out_stream, func, lock, finished, crashed, keep_order):
+
+            async def set_finish():
+                active_tasks.pop()
+                if not active_tasks:
+                    # This thread is the last active one
+                    # for the current transformer.
+                    await out_stream.put_end()
+                finished.set()
+
+            Future = asyncio.Future
+            while True:
                 async with lock:
-                    # This locked block ensures that
-                    # input is read in order and their corresponding
-                    # result placeholders (Future objects) are
-                    # put in the output stream in order.
                     if finished.is_set():
+                        await set_finish()
                         return
-                    if self._err:
-                        return
-                    try:
-                        x = await in_stream.__anext__()
-                        fut = asyncio.Future()
-                        await out_stream.put(fut)
-                    except StopAsyncIteration:
-                        finished.set()
-                        await out_stream.put_end()
-                        return
-                    except Exception as e:
-                        finished.set()
-                        self._err.append(e)
+                    if crashed.is_set():
+                        await set_finish()
                         return
 
-                try:
-                    y = await func(x)
-                    fut.set_result(y)
-                except Exception as e:
-                    if return_exceptions:
-                        fut.set_result(e)
-                    else:
-                        fut.set_exception(e)
-                        finished.set()
+                    try:
+                        x = await in_stream.__anext__()
+                    except StopAsyncIteration:
+                        await set_finish()
                         return
+                    except Exception as e:
+                        self._upstream_err.append(e)
+                        await set_finish()
+                        return
+                    else:
+                        if keep_order:
+                            fut = Future()
+                            await out_stream.put(fut)
+
+                if keep_order:
+                    try:
+                        y = await func(x)
+                    except Exception as e:
+                        if self.return_exceptions:
+                            y = e
+                        else:
+                            fut.set_exception(e)
+                            self._crashed.set()
+                            await set_finish()
+                            return
+                    fut.set_result(y)
+                else:
+                    try:
+                        y = await func(x)
+                    except Exception as e:
+                        await out_stream.put(e)
+                        if not self.return_exceptions:
+                            self._crashed.set()
+                            await set_finish()
+                            return
+                    else:
+                        await out_stream.put(y)
 
         lock = asyncio.Lock()
         finished = asyncio.Event()
 
+        active_tasks = list(range(self.workers))
         self._tasks = [
             asyncio.create_task(
                 _process(
@@ -444,14 +462,23 @@ class ConcurrentTransformer(Stream):
                     self.func,
                     lock,
                     finished,
-                    self.return_exceptions),
+                    self._crashed,
+                    self.keep_order),
                 name=f'transformer-{i}',
             )
             for i in range(self.workers)
         ]
 
     def _start_sync(self):
-        async def _put_input_in_queue(q_in, q_out):
+        async def _put_input_in_queue(q_in, q_out, crashed, finished):
+            async def put_end():
+                while True:
+                    try:
+                        q_out.put_end(block=False)
+                        return
+                    except queue.Full:
+                        await asyncio.sleep(q_out.PUT_SLEEP)
+
             try:
                 async for x in q_in:
                     while True:
@@ -459,21 +486,27 @@ class ConcurrentTransformer(Stream):
                             q_out.put(x, block=False)
                             break
                         except queue.Full:
-                            await asyncio.sleep(q_out.PUT_SLEEP)
-                while True:
-                    try:
-                        q_out.put_end(block=False)
-                        break
-                    except queue.Full:
+                            if crashed.is_set() or finished.is_set():
+                                await put_end()
+                                return
                         await asyncio.sleep(q_out.PUT_SLEEP)
             except Exception as e:
-                self._err.append(e)
+                self._upstream_err.append(e)
+            await put_end()
 
-        q_in = _sync_streamer.IterQueue(self.workers * 8, self._to_shutdown)
-        t1 = asyncio.create_task(_put_input_in_queue(self._instream, q_in))
+        q_in = _sync_streamer.IterQueue(1024, self._crashed)
+        finished = threading.Event()
+        t1 = asyncio.create_task(_put_input_in_queue(self._instream, q_in, self._crashed, finished))
         t2 = _sync_streamer.transform(
-            q_in, self._outstream, self.func,
-            self.workers, self.return_exceptions, self._err)
+            q_in,
+            self._outstream,
+            self.func,
+            workers=self.workers,
+            return_exceptions=self.return_exceptions,
+            keep_order=self.keep_order,
+            crashed=self._crashed,
+            upstream_err=self._upstream_err,
+        )
         self._tasks = [t1] + t2
 
     async def _astop(self):
@@ -489,16 +522,30 @@ class ConcurrentTransformer(Stream):
         self._tasks = []
 
     async def _get_next(self):
+        if self._upstream_err:
+            raise self._upstream_err[0]
         try:
-            if self._err:
-                raise self._err[0]
-            fut = await self._outstream.__anext__()
-            if self._async:
-                return await fut
-            else:
-                while not fut.done():
-                    await asyncio.sleep(IterQueue.GET_SLEEP)
-                return fut.result()
-        except:
-            await self._astop()
+            y = await self._outstream.__anext__()
+        except StopAsyncIteration:
+            if self._upstream_err:
+                raise self._upstream_err[0]
             raise
+        if self.keep_order:
+            if self._async:
+                try:
+                    return await y
+                except Exception:
+                    self._crashed.set()
+                    raise
+            else:
+                while not y.done():
+                    await asyncio.sleep(IterQueue.GET_SLEEP)
+                try:
+                    return y.result()
+                except Exception:
+                    self._crashed.set()
+                    raise
+        if isinstance(y, Exception) and not self.return_exceptions:
+            self._crashed.set()
+            raise y
+        return y
